"""RGB representation. Simply inherit FakeRepresentation that does all this needs (i.e. copy pasta the frames)"""
from .fake_representation import FakeRepresentation

class RGB(FakeRepresentation): pass # pylint: disable=missing-class-docstring, multiple-statements
