"""init file"""
from .stored_represntation import StoredRepresentation, NpzRepresentation
from .normed_representation import NormedRepresentation
from .task_mapper import TaskMapper
from .cv_representations import (RGBRepresentation, HSVRepresentation, EdgesRepresentation, SemanticRepresentation,
                                 DepthRepresentation, OpticalFlowRepresentation, NormalsRepresentation,
                                 ColorRepresentation)
